@@ -3,6 +3,8 @@
 
 """Determining whether files are being measured/reported or not."""
 
+from __future__ import annotations
+
 import importlib.util
 import inspect
 import itertools
@@ -13,33 +15,49 @@ import sys
 import sysconfig
 import traceback
 
+from types import FrameType, ModuleType
+from typing import (
+    cast, Any, TYPE_CHECKING,
+)
+from collections.abc import Iterable
+
 from coverage import env
 from coverage.disposition import FileDisposition, disposition_init
 from coverage.exceptions import CoverageException, PluginError
-from coverage.files import TreeMatcher, FnmatchMatcher, ModuleMatcher
+from coverage.files import TreeMatcher, GlobMatcher, ModuleMatcher
 from coverage.files import prep_patterns, find_python_files, canonical_filename
 from coverage.misc import sys_modules_saved
 from coverage.python import source_for_file, source_for_morf
+from coverage.types import TFileDisposition, TMorf, TWarnFn, TDebugCtl
+
+if TYPE_CHECKING:
+    from coverage.config import CoverageConfig
+    from coverage.plugin_support import Plugins
 
 
 # Pypy has some unusual stuff in the "stdlib".  Consider those locations
 # when deciding where the stdlib is.  These modules are not used for anything,
 # they are modules importable from the pypy lib directories, so that we can
 # find those directories.
-_structseq = _pypy_irc_topic = None
+modules_we_happen_to_have: list[ModuleType] = [
+    inspect, itertools, os, platform, re, sysconfig, traceback,
+]
+
 if env.PYPY:
     try:
         import _structseq
+        modules_we_happen_to_have.append(_structseq)
     except ImportError:
         pass
 
     try:
         import _pypy_irc_topic
+        modules_we_happen_to_have.append(_pypy_irc_topic)
     except ImportError:
         pass
 
 
-def canonical_path(morf, directory=False):
+def canonical_path(morf: TMorf, directory: bool = False) -> str:
     """Return the canonical path of the module or file `morf`.
 
     If the module is a package, then return its directory. If it is a
@@ -53,7 +71,7 @@ def canonical_path(morf, directory=False):
     return morf_path
 
 
-def name_for_module(filename, frame):
+def name_for_module(filename: str, frame: FrameType | None) -> str:
     """Get the name of the module for a filename and frame.
 
     For configurability's sake, we allow __main__ modules to be matched by
@@ -66,24 +84,16 @@ def name_for_module(filename, frame):
 
     """
     module_globals = frame.f_globals if frame is not None else {}
-    if module_globals is None:          # pragma: only ironpython
-        # IronPython doesn't provide globals: https://github.com/IronLanguages/main/issues/1296
-        module_globals = {}
+    dunder_name: str = module_globals.get("__name__", None)
 
-    dunder_name = module_globals.get('__name__', None)
-
-    if isinstance(dunder_name, str) and dunder_name != '__main__':
+    if isinstance(dunder_name, str) and dunder_name != "__main__":
         # This is the usual case: an imported module.
         return dunder_name
 
-    loader = module_globals.get('__loader__', None)
-    for attrname in ('fullname', 'name'):   # attribute renamed in py3.2
-        if hasattr(loader, attrname):
-            fullname = getattr(loader, attrname)
-        else:
-            continue
-
-        if isinstance(fullname, str) and fullname != '__main__':
+    spec = module_globals.get("__spec__", None)
+    if spec:
+        fullname = spec.name
+        if isinstance(fullname, str) and fullname != "__main__":
             # Module loaded via: runpy -m
             return fullname
 
@@ -95,20 +105,20 @@ def name_for_module(filename, frame):
         return dunder_name
 
 
-def module_is_namespace(mod):
+def module_is_namespace(mod: ModuleType) -> bool:
     """Is the module object `mod` a PEP420 namespace module?"""
-    return hasattr(mod, '__path__') and getattr(mod, '__file__', None) is None
+    return hasattr(mod, "__path__") and getattr(mod, "__file__", None) is None
 
 
-def module_has_file(mod):
+def module_has_file(mod: ModuleType) -> bool:
     """Does the module object `mod` have an existing __file__ ?"""
-    mod__file__ = getattr(mod, '__file__', None)
+    mod__file__ = getattr(mod, "__file__", None)
     if mod__file__ is None:
         return False
     return os.path.exists(mod__file__)
 
 
-def file_and_path_for_module(modulename):
+def file_and_path_for_module(modulename: str) -> tuple[str | None, list[str]]:
     """Find the file and search path for `modulename`.
 
     Returns:
@@ -124,38 +134,24 @@ def file_and_path_for_module(modulename):
         pass
     else:
         if spec is not None:
-            if spec.origin != "namespace":
-                filename = spec.origin
+            filename = spec.origin
             path = list(spec.submodule_search_locations or ())
     return filename, path
 
 
-def add_stdlib_paths(paths):
+def add_stdlib_paths(paths: set[str]) -> None:
     """Add paths where the stdlib can be found to the set `paths`."""
     # Look at where some standard modules are located. That's the
     # indication for "installed with the interpreter". In some
     # environments (virtualenv, for example), these modules may be
     # spread across a few locations. Look at all the candidate modules
     # we've imported, and take all the different ones.
-    modules_we_happen_to_have = [
-        inspect, itertools, os, platform, re, sysconfig, traceback,
-        _pypy_irc_topic, _structseq,
-    ]
     for m in modules_we_happen_to_have:
-        if m is not None and hasattr(m, "__file__"):
+        if hasattr(m, "__file__"):
             paths.add(canonical_path(m, directory=True))
 
-    if _structseq and not hasattr(_structseq, '__file__'):
-        # PyPy 2.4 has no __file__ in the builtin modules, but the code
-        # objects still have the file names.  So dig into one to find
-        # the path to exclude.  The "filename" might be synthetic,
-        # don't be fooled by those.
-        structseq_file = _structseq.structseq_new.__code__.co_filename
-        if not structseq_file.startswith("<"):
-            paths.add(canonical_path(structseq_file))
 
-
-def add_third_party_paths(paths):
+def add_third_party_paths(paths: set[str]) -> None:
     """Add locations for third-party packages to the set `paths`."""
     # Get the paths that sysconfig knows about.
     scheme_names = set(sysconfig.get_scheme_names())
@@ -169,7 +165,7 @@ def add_third_party_paths(paths):
                 paths.add(config_paths[path_name])
 
 
-def add_coverage_paths(paths):
+def add_coverage_paths(paths: set[str]) -> None:
     """Add paths where coverage.py code can be found to the set `paths`."""
     cover_path = canonical_path(__file__, directory=True)
     paths.add(cover_path)
@@ -177,43 +173,23 @@ def add_coverage_paths(paths):
         # Don't include our own test code.
         paths.add(os.path.join(cover_path, "tests"))
 
-        # When testing, we use PyContracts, which should be considered
-        # part of coverage.py, and it uses six. Exclude those directories
-        # just as we exclude ourselves.
-        if env.USE_CONTRACTS:
-            import contracts
-            import six
-            for mod in [contracts, six]:
-                paths.add(canonical_path(mod))
-
 
 class InOrOut:
     """Machinery for determining what files to measure."""
 
-    def __init__(self, warn, debug):
+    def __init__(
+        self,
+        config: CoverageConfig,
+        warn: TWarnFn,
+        debug: TDebugCtl | None,
+        include_namespace_packages: bool,
+    ) -> None:
         self.warn = warn
         self.debug = debug
+        self.include_namespace_packages = include_namespace_packages
 
-        # The matchers for should_trace.
-        self.source_match = None
-        self.source_pkgs_match = None
-        self.pylib_paths = self.cover_paths = self.third_paths = None
-        self.pylib_match = self.cover_match = self.third_match = None
-        self.include_match = self.omit_match = None
-        self.plugins = []
-        self.disp_class = FileDisposition
-
-        # The source argument can be directories or package names.
-        self.source = []
-        self.source_pkgs = []
-        self.source_pkgs_unmatched = []
-        self.omit = self.include = None
-
-        # Is the source inside a third-party area?
-        self.source_in_third = False
-
-    def configure(self, config):
-        """Apply the configuration to get ready for decision-time."""
+        self.source: list[str] = []
+        self.source_pkgs: list[str] = []
         self.source_pkgs.extend(config.source_pkgs)
         for src in config.source or []:
             if os.path.isdir(src):
@@ -222,31 +198,38 @@ class InOrOut:
                 self.source_pkgs.append(src)
         self.source_pkgs_unmatched = self.source_pkgs[:]
 
-        self.omit = prep_patterns(config.run_omit)
         self.include = prep_patterns(config.run_include)
+        self.omit = prep_patterns(config.run_omit)
 
         # The directories for files considered "installed with the interpreter".
-        self.pylib_paths = set()
+        self.pylib_paths: set[str] = set()
         if not config.cover_pylib:
             add_stdlib_paths(self.pylib_paths)
 
         # To avoid tracing the coverage.py code itself, we skip anything
         # located where we are.
-        self.cover_paths = set()
+        self.cover_paths: set[str] = set()
         add_coverage_paths(self.cover_paths)
 
         # Find where third-party packages are installed.
-        self.third_paths = set()
+        self.third_paths: set[str] = set()
         add_third_party_paths(self.third_paths)
 
-        def debug(msg):
+        def _debug(msg: str) -> None:
             if self.debug:
                 self.debug.write(msg)
 
+        # The matchers for should_trace.
+
         # Generally useful information
-        debug("sys.path:" + "".join(f"\n    {p}" for p in sys.path))
+        _debug("sys.path:" + "".join(f"\n    {p}" for p in sys.path))
 
         # Create the matchers we need for should_trace
+        self.source_match = None
+        self.source_pkgs_match = None
+        self.pylib_match = None
+        self.include_match = self.omit_match = None
+
         if self.source or self.source_pkgs:
             against = []
             if self.source:
@@ -255,55 +238,61 @@ class InOrOut:
             if self.source_pkgs:
                 self.source_pkgs_match = ModuleMatcher(self.source_pkgs, "source_pkgs")
                 against.append(f"modules {self.source_pkgs_match!r}")
-            debug("Source matching against " + " and ".join(against))
+            _debug("Source matching against " + " and ".join(against))
         else:
             if self.pylib_paths:
                 self.pylib_match = TreeMatcher(self.pylib_paths, "pylib")
-                debug(f"Python stdlib matching: {self.pylib_match!r}")
+                _debug(f"Python stdlib matching: {self.pylib_match!r}")
         if self.include:
-            self.include_match = FnmatchMatcher(self.include, "include")
-            debug(f"Include matching: {self.include_match!r}")
+            self.include_match = GlobMatcher(self.include, "include")
+            _debug(f"Include matching: {self.include_match!r}")
         if self.omit:
-            self.omit_match = FnmatchMatcher(self.omit, "omit")
-            debug(f"Omit matching: {self.omit_match!r}")
+            self.omit_match = GlobMatcher(self.omit, "omit")
+            _debug(f"Omit matching: {self.omit_match!r}")
 
         self.cover_match = TreeMatcher(self.cover_paths, "coverage")
-        debug(f"Coverage code matching: {self.cover_match!r}")
+        _debug(f"Coverage code matching: {self.cover_match!r}")
 
         self.third_match = TreeMatcher(self.third_paths, "third")
-        debug(f"Third-party lib matching: {self.third_match!r}")
+        _debug(f"Third-party lib matching: {self.third_match!r}")
 
         # Check if the source we want to measure has been installed as a
         # third-party package.
+        # Is the source inside a third-party area?
+        self.source_in_third_paths = set()
         with sys_modules_saved():
             for pkg in self.source_pkgs:
                 try:
                     modfile, path = file_and_path_for_module(pkg)
-                    debug(f"Imported source package {pkg!r} as {modfile!r}")
+                    _debug(f"Imported source package {pkg!r} as {modfile!r}")
                 except CoverageException as exc:
-                    debug(f"Couldn't import source package {pkg!r}: {exc}")
+                    _debug(f"Couldn't import source package {pkg!r}: {exc}")
                     continue
                 if modfile:
                     if self.third_match.match(modfile):
-                        debug(
-                            f"Source is in third-party because of source_pkg {pkg!r} at {modfile!r}"
+                        _debug(
+                            f"Source in third-party: source_pkg {pkg!r} at {modfile!r}",
                         )
-                        self.source_in_third = True
+                        self.source_in_third_paths.add(canonical_path(source_for_file(modfile)))
                 else:
                     for pathdir in path:
                         if self.third_match.match(pathdir):
-                            debug(
-                                f"Source is in third-party because of {pkg!r} path directory " +
-                                f"at {pathdir!r}"
+                            _debug(
+                                f"Source in third-party: {pkg!r} path directory at {pathdir!r}",
                             )
-                            self.source_in_third = True
+                            self.source_in_third_paths.add(pathdir)
 
         for src in self.source:
             if self.third_match.match(src):
-                debug(f"Source is in third-party because of source directory {src!r}")
-                self.source_in_third = True
+                _debug(f"Source in third-party: source directory {src!r}")
+                self.source_in_third_paths.add(src)
+        self.source_in_third_match = TreeMatcher(self.source_in_third_paths, "source_in_third")
+        _debug(f"Source in third-party matching: {self.source_in_third_match}")
 
-    def should_trace(self, filename, frame=None):
+        self.plugins: Plugins
+        self.disp_class: type[TFileDisposition] = FileDisposition
+
+    def should_trace(self, filename: str, frame: FrameType | None = None) -> TFileDisposition:
         """Decide whether to trace execution in `filename`, with a reason.
 
         This function is called from the trace function.  As each new file name
@@ -315,14 +304,14 @@ class InOrOut:
         original_filename = filename
         disp = disposition_init(self.disp_class, filename)
 
-        def nope(disp, reason):
+        def nope(disp: TFileDisposition, reason: str) -> TFileDisposition:
             """Simple helper to make it easy to return NO."""
             disp.trace = False
             disp.reason = reason
             return disp
 
-        if original_filename.startswith('<'):
-            return nope(disp, "not a real original file name")
+        if original_filename.startswith("<"):
+            return nope(disp, "original file name is not real")
 
         if frame is not None:
             # Compiled Python files have two file names: frame.f_code.co_filename is
@@ -331,10 +320,11 @@ class InOrOut:
             # .pyc files can be moved after compilation (for example, by being
             # installed), we look for __file__ in the frame and prefer it to the
             # co_filename value.
-            dunder_file = frame.f_globals and frame.f_globals.get('__file__')
+            dunder_file = frame.f_globals and frame.f_globals.get("__file__")
             if dunder_file:
-                filename = source_for_file(dunder_file)
-                if original_filename and not original_filename.startswith('<'):
+                # Danger: __file__ can (rarely?) be of type Path.
+                filename = source_for_file(str(dunder_file))
+                if original_filename and not original_filename.startswith("<"):
                     orig = os.path.basename(original_filename)
                     if orig != os.path.basename(filename):
                         # Files shouldn't be renamed when moved. This happens when
@@ -346,19 +336,15 @@ class InOrOut:
             # Empty string is pretty useless.
             return nope(disp, "empty string isn't a file name")
 
-        if filename.startswith('memory:'):
+        if filename.startswith("memory:"):
             return nope(disp, "memory isn't traceable")
 
-        if filename.startswith('<'):
+        if filename.startswith("<"):
             # Lots of non-file execution is represented with artificial
             # file names like "<string>", "<doctest readme.txt[0]>", or
             # "<exec_function>".  Don't ever trace these executions, since we
             # can't do anything with the data later anyway.
-            return nope(disp, "not a real file name")
-
-        # Jython reports the .class file to the tracer, use the source file.
-        if filename.endswith("$py.class"):
-            filename = filename[:-9] + ".py"
+            return nope(disp, "file name is not real")
 
         canonical = canonical_filename(filename)
         disp.canonical_filename = canonical
@@ -379,7 +365,7 @@ class InOrOut:
                         disp.has_dynamic_filename = True
                     else:
                         disp.source_filename = canonical_filename(
-                            file_tracer.source_filename()
+                            file_tracer.source_filename(),
                         )
                     break
             except Exception:
@@ -396,7 +382,7 @@ class InOrOut:
         if not disp.has_dynamic_filename:
             if not disp.source_filename:
                 raise PluginError(
-                    f"Plugin {plugin!r} didn't set source_filename for '{disp.original_filename}'"
+                    f"Plugin {plugin!r} didn't set source_filename for '{disp.original_filename}'",
                 )
             reason = self.check_include_omit_etc(disp.source_filename, frame)
             if reason:
@@ -404,7 +390,7 @@ class InOrOut:
 
         return disp
 
-    def check_include_omit_etc(self, filename, frame):
+    def check_include_omit_etc(self, filename: str, frame: FrameType | None) -> str | None:
         """Check a file name against the include, omit, etc, rules.
 
         Returns a string or None.  String means, don't trace, and is the reason
@@ -432,9 +418,8 @@ class InOrOut:
                     ok = True
             if not ok:
                 return extra + "falls outside the --source spec"
-            if not self.source_in_third:
-                if self.third_match.match(filename):
-                    return "inside --source, but is third-party"
+            if self.third_match.match(filename) and not self.source_in_third_match.match(filename):
+                return "inside --source, but is third-party"
         elif self.include_match:
             if not self.include_match.match(filename):
                 return "falls outside the --include trees"
@@ -466,13 +451,13 @@ class InOrOut:
         # No reason found to skip this file.
         return None
 
-    def warn_conflicting_settings(self):
+    def warn_conflicting_settings(self) -> None:
         """Warn if there are settings that conflict."""
         if self.include:
             if self.source or self.source_pkgs:
                 self.warn("--include is ignored because --source is set", slug="include-ignored")
 
-    def warn_already_imported_files(self):
+    def warn_already_imported_files(self) -> None:
         """Warn if files have already been imported that we will be measuring."""
         if self.include or self.source or self.source_pkgs:
             warned = set()
@@ -497,19 +482,19 @@ class InOrOut:
                     msg = f"Already imported a file that will be measured: {filename}"
                     self.warn(msg, slug="already-imported")
                     warned.add(filename)
-                elif self.debug and self.debug.should('trace'):
+                elif self.debug and self.debug.should("trace"):
                     self.debug.write(
                         "Didn't trace already imported file {!r}: {}".format(
-                            disp.original_filename, disp.reason
-                        )
+                            disp.original_filename, disp.reason,
+                        ),
                     )
 
-    def warn_unimported_source(self):
+    def warn_unimported_source(self) -> None:
         """Warn about source packages that were of interest, but never traced."""
         for pkg in self.source_pkgs_unmatched:
             self._warn_about_unmeasured_code(pkg)
 
-    def _warn_about_unmeasured_code(self, pkg):
+    def _warn_about_unmeasured_code(self, pkg: str) -> None:
         """Warn about a package or module that we never traced.
 
         `pkg` is a string, the name of the package or module.
@@ -535,28 +520,28 @@ class InOrOut:
         msg = f"Module {pkg} was previously imported, but not measured"
         self.warn(msg, slug="module-not-measured")
 
-    def find_possibly_unexecuted_files(self):
+    def find_possibly_unexecuted_files(self) -> Iterable[tuple[str, str | None]]:
         """Find files in the areas of interest that might be untraced.
 
         Yields pairs: file path, and responsible plug-in name.
         """
         for pkg in self.source_pkgs:
-            if (not pkg in sys.modules or
+            if (pkg not in sys.modules or
                 not module_has_file(sys.modules[pkg])):
                 continue
-            pkg_file = source_for_file(sys.modules[pkg].__file__)
+            pkg_file = source_for_file(cast(str, sys.modules[pkg].__file__))
             yield from self._find_executable_files(canonical_path(pkg_file))
 
         for src in self.source:
             yield from self._find_executable_files(src)
 
-    def _find_plugin_files(self, src_dir):
+    def _find_plugin_files(self, src_dir: str) -> Iterable[tuple[str, str]]:
         """Get executable files from the plugins."""
         for plugin in self.plugins.file_tracers:
             for x_file in plugin.find_executable_files(src_dir):
                 yield x_file, plugin._coverage_plugin_name
 
-    def _find_executable_files(self, src_dir):
+    def _find_executable_files(self, src_dir: str) -> Iterable[tuple[str, str | None]]:
         """Find executable files in `src_dir`.
 
         Search for files in `src_dir` that can be executed because they
@@ -566,18 +551,21 @@ class InOrOut:
         Yield the file path, and the plugin name that handles the file.
 
         """
-        py_files = ((py_file, None) for py_file in find_python_files(src_dir))
+        py_files = (
+            (py_file, None) for py_file in
+            find_python_files(src_dir, self.include_namespace_packages)
+        )
         plugin_files = self._find_plugin_files(src_dir)
 
         for file_path, plugin_name in itertools.chain(py_files, plugin_files):
             file_path = canonical_filename(file_path)
             if self.omit_match and self.omit_match.match(file_path):
                 # Turns out this file was omitted, so don't pull it back
-                # in as unexecuted.
+                # in as un-executed.
                 continue
             yield file_path, plugin_name
 
-    def sys_info(self):
+    def sys_info(self) -> Iterable[tuple[str, Any]]:
         """Our information for Coverage.sys_info.
 
         Returns a list of (key, value) pairs.
@@ -586,20 +574,21 @@ class InOrOut:
             ("coverage_paths", self.cover_paths),
             ("stdlib_paths", self.pylib_paths),
             ("third_party_paths", self.third_paths),
+            ("source_in_third_party_paths", self.source_in_third_paths),
         ]
 
         matcher_names = [
-            'source_match', 'source_pkgs_match',
-            'include_match', 'omit_match',
-            'cover_match', 'pylib_match', 'third_match',
-            ]
+            "source_match", "source_pkgs_match",
+            "include_match", "omit_match",
+            "cover_match", "pylib_match", "third_match", "source_in_third_match",
+        ]
 
         for matcher_name in matcher_names:
             matcher = getattr(self, matcher_name)
             if matcher:
                 matcher_info = matcher.info()
             else:
-                matcher_info = '-none-'
+                matcher_info = "-none-"
             info.append((matcher_name, matcher_info))
 
         return info
